using System;
using UnityEngine;
using System.Collections.Generic;

#if GOOGLE_MOBILE_ADS
using GoogleMobileAds.Api;
using GoogleMobileAds.Common;
using System.Collections;

#endif

public class GoogleAdMobController : MonoBehaviour
{
#if GOOGLE_MOBILE_ADS
    private AppOpenAd appOpenAd;
    private BannerView _bannerView;
    public InterstitialAd _interstitialAd;
    private RewardedAd _rewardedAd;
    private RewardedInterstitialAd _rewardedInterstitialAd;

    AdSize adSize;
    AdPosition adPosition;

#endif
    private bool isShowingAppOpenAd;

    #region UNITY MONOBEHAVIOR METHODS

    private void OnEnable()
    {
        Initialize();
    }
    
    public void Initialize()
    {
#if GOOGLE_MOBILE_ADS

        MobileAds.SetiOSAppPauseOnBackground(true);

        List<String> deviceIds = new List<String>() { AdRequest.TestDeviceSimulator };
        MobileAds.SetRequestConfiguration(new RequestConfiguration
        {
            TestDeviceIds = deviceIds,
            MaxAdContentRating = MaxAdContentRating.G,
            TagForChildDirectedTreatment = TagForChildDirectedTreatment.True

        });
        MobileAds.RaiseAdEventsOnUnityMainThread = true;
        MobileAds.Initialize(HandleInitCompleteAction);
    }

    private void HandleInitCompleteAction(InitializationStatus initstatus)
    {
        MobileAdsEventExecutor.ExecuteInUpdate(() =>
        {
            RequestBannerAd(AdPosition.Bottom, true);
        });
#endif
    }

    #endregion

    #region HELPER METHODS

#if GOOGLE_MOBILE_ADS
    private AdRequest CreateAdRequest() { return new AdRequest(); }
#endif
    #endregion

    #region BANNER ADS


#if GOOGLE_MOBILE_ADS
    public void RequestBannerAd(AdPosition bannerPosition, bool isAdaptive)
    {
#if GOOGLE_MOBILE_ADS
        // Test Ads
        string adUnitId = "unused";
#if UNITY_EDITOR
        adUnitId = "unused";
#elif UNITY_ANDROID
        adUnitId = "ca-app-pub-3940256099942544/6300978111";
#elif UNITY_IPHONE
        adUnitId = "ca-app-pub-3940256099942544/2934735716";
#else
        adUnitId = "unexpected_platform";
#endif

        CreateBannerView(isAdaptive, adUnitId, bannerPosition);
        ListenToBannerAdEvents();

        _bannerView.LoadAd(CreateAdRequest());

        print("Request Banner Ad");

#endif
    }

    /// <summary>
    /// Creates a 320x50 banner view at top of the screen.
    /// </summary>
    public void CreateBannerView(bool isAdaptive, string _adUnitId, AdPosition _scenename)
    {
        Debug.Log("Creating banner view");

        // If we already have a banner, destroy the old one.
        if (_bannerView != null)
        {
            DestroyBannerAd();
        }

        if (isAdaptive)
        {
            adSize = AdSize.GetLandscapeAnchoredAdaptiveBannerAdSizeWithWidth(AdSize.FullWidth);
        }
        else
        {
            adSize = AdSize.Banner;
        }
        
        _bannerView = new BannerView(_adUnitId, adSize, _scenename);
    }

    /// <summary>
    /// listen to events the banner view may raise.
    /// </summary>
    private void ListenToBannerAdEvents()
    {
        // Raised when an ad is loaded into the banner view.
        _bannerView.OnBannerAdLoaded += () =>
        {
            Debug.Log("Banner view loaded an ad with response : "
                + _bannerView.GetResponseInfo());
        };
        // Raised when an ad fails to load into the banner view.
        _bannerView.OnBannerAdLoadFailed += (LoadAdError error) =>
        {
            Debug.LogError("Banner view failed to load an ad with error : "
                + error);
        };
        // Raised when the ad is estimated to have earned money.
        _bannerView.OnAdPaid += (AdValue adValue) =>
        {
            Debug.Log(String.Format("Banner view paid {0} {1}.",
                adValue.Value,
                adValue.CurrencyCode));
            
        };
        // Raised when an impression is recorded for an ad.
        _bannerView.OnAdImpressionRecorded += () =>
        {
            Debug.Log("Banner view recorded an impression.");
        };
        // Raised when a click is recorded for an ad.
        _bannerView.OnAdClicked += () =>
        {
            Debug.Log("Banner view was clicked.");
        };
        // Raised when an ad opened full screen content.
        _bannerView.OnAdFullScreenContentOpened += () =>
        {
            Debug.Log("Banner view full screen content opened.");
        };
        // Raised when the ad closed full screen content.
        _bannerView.OnAdFullScreenContentClosed += () =>
        {
            Debug.Log("Banner view full screen content closed.");
        };
    }

#endif


    public void DestroyBannerAd()
    {
#if GOOGLE_MOBILE_ADS
        if (_bannerView != null)
        {
            _bannerView.Destroy();
            _bannerView = null;
        }
#endif
    }
    public void HideBannerAd()
    {
#if GOOGLE_MOBILE_ADS
        if (_bannerView != null) _bannerView.Hide();
#endif
    }
    #endregion
}